/*
Module : CriticalSectionEx.h
Purpose: Defines the interface for an MFC class which provides a critical section with a spin count. The 
         class is based in part on the code presented in Chapter Ten of the book "Programming Applications 
         for Microsoft Windows" by Jeffrey Richter. 
Created: PJN / 17-01-2016
History: PJN / 08-12-2018 1. Updated copyright details
                          2. Removed deprecated TSafe.h from download
         PJN / 20-12-2019 1. Updated copyright details.
                          2. Fixed various Clang-Tidy static code analysis warnings in the code.

Copyright (c) 2016 - 2020 by PJ Naughter. (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 
*/


/////////////////////////////// Macros / Defines //////////////////////////////

#pragma once

#ifndef __CRITICALSECTIONEX_H__
#define __CRITICALSECTIONEX_H__

#ifndef __AFXMT_H__
#include <afxmt.h>
#pragma message("To avoid this message, please put afxmt.h in your pre compiled header (normally stdafx.h)")
#endif //#ifndef __AFXMT_H__

#ifndef CCRITICALSECTIONEX_EXT_CLASS
#define CCRITICALSECTIONEX_EXT_CLASS
#endif //#ifndef CCRITICALSECTIONEX_EXT_CLASS


/////////////////////////////// Classes ///////////////////////////////////////

//Extended version of MFC's CCriticalSection class. This version provides timeouts and spin counts
class CCRITICALSECTIONEX_EXT_CLASS CCriticalSectionEx : public CSyncObject
{
public:
//Constructor / Destructors
#pragma warning(suppress: 26455)
  CCriticalSectionEx(_In_ DWORD dwSpinCount = 1000) : CSyncObject(nullptr)
  {
    if (!InitializeCriticalSectionAndSpinCount(&m_sect, dwSpinCount))
      AtlThrow(HRESULT_FROM_WIN32(GetLastError()));
  }

  CCriticalSectionEx(const CCriticalSectionEx&) = delete;
  CCriticalSectionEx(CCriticalSectionEx&&) = delete;

  ~CCriticalSectionEx() //NOLINT(modernize-use-override)
  {
    //Free up the critical section
    DeleteCriticalSection(&m_sect);
  }

//Methods
  CCriticalSectionEx& operator=(const CCriticalSectionEx&) = delete;
  CCriticalSectionEx& operator=(CCriticalSectionEx&&) = delete;

  operator CRITICAL_SECTION*() noexcept
  {
    return &m_sect;
  }

  _When_(return != 0, _Acquires_lock_(this->m_sect)) BOOL Lock() noexcept
  {
    EnterCriticalSection(&m_sect);
    return TRUE;
  }

  _Releases_lock_(this->m_sect) BOOL Unlock() noexcept override
  {
    LeaveCriticalSection(&m_sect);
    return TRUE;
  }

  #pragma warning(suppress: 28204)
  _When_(return != 0, _Acquires_lock_(this->m_sect)) BOOL Lock(_In_ DWORD dwTimeout) noexcept override
  {
    //If any timeout is specified (other than INFINITE) then try to acquire
    //the critical section using "TryEnterCriticalSection". Note that calling this
    //function is effectively like waiting 0 milliseconds for the critical section.
    //In other words currently the only usable values for dwTimeout are 0 and INFINITE.
    if (dwTimeout != INFINITE)
      return TryEnterCriticalSection(&m_sect);
    else
    {
      EnterCriticalSection(&m_sect);
      return TRUE;
    }
  }

  DWORD SetSpinCount(_In_ DWORD dwSpinCount) noexcept
  {
    return SetCriticalSectionSpinCount(&m_sect, dwSpinCount);
  }

//Member variables
  CRITICAL_SECTION m_sect;
};

#endif //#ifndef __CRITICALSECTIONEX_H__
